from __future__ import generator_stop
import math

import keras
import numpy as np


class Distributor(keras.utils.Sequence):

    def __init__(self, generator: keras.utils.Sequence, model: keras.models, batch_size: int = 256,
                 fit_call: bool = True):
        self.generator = generator
        self.model = model
        self.batch_size = batch_size
        self.fit_call = fit_call

    def _get_model_rank(self):
        mod_out = self.model.output_shape
        if isinstance(mod_out, tuple):
            # only one output branch: (None, ahead)
            mod_rank = 1
        elif isinstance(mod_out, list):
            # multiple output branches, e.g.: [(None, ahead), (None, ahead)]
            mod_rank = len(mod_out)
        else:  # pragma: no cover
            raise TypeError("model output shape must either be tuple or list.")
        return mod_rank

    def _get_number_of_mini_batches(self, values):
        return math.ceil(values[0].shape[0] / self.batch_size)

    def distribute_on_batches(self, fit_call=True):
        while True:
            for k, v in enumerate(self.generator):
                # get rank of output
                mod_rank = self._get_model_rank()
                # get number of mini batches
                num_mini_batches = self._get_number_of_mini_batches(v)
                x_total = np.copy(v[0])
                y_total = np.copy(v[1])
                for prev, curr in enumerate(range(1, num_mini_batches+1)):
                    x = x_total[prev*self.batch_size:curr*self.batch_size, ...]
                    y = [y_total[prev*self.batch_size:curr*self.batch_size, ...] for _ in range(mod_rank)]
                    if x is not None:
                        yield (x, y)
                        if (k + 1) == len(self.generator) and curr == num_mini_batches and not fit_call:
                            return

    def __len__(self):
        num_batch = 0
        for _ in self.distribute_on_batches(fit_call=False):
            num_batch += 1
        return num_batch
