
__author__ = "Lukas Leufen, Felix Kleinert"
__date__ = '2019-12-05'

import logging
import os
import json

from src.modules.run_environment import RunEnvironment
from src.data_handling.data_distributor import Distributor


class Training(RunEnvironment):

    def __init__(self):
        super().__init__()
        self.model = self.data_store.get("model", "general.model")
        self.train_set = None
        self.val_set = None
        self.test_set = None
        self.batch_size = self.data_store.get("batch_size", "general.model")
        self.epochs = self.data_store.get("epochs", "general.model")
        self.checkpoint = self.data_store.get("checkpoint", "general.model")
        self.lr_sc = self.data_store.get("lr_decay", "general.model")
        self.experiment_name = self.data_store.get("experiment_name", "general")
        self._run()

    def _run(self):
        self.set_generators()
        self.make_predict_function()
        self.train()

    def make_predict_function(self):
        # create the predict function before distributing. This is necessary, because tf will compile the predict
        # function just in the moment it is used the first time. This can cause problems, if the model is distributed
        # on different workers. To prevent this, the function is pre-compiled. See discussion @
        # https://stackoverflow.com/questions/40850089/is-keras-thread-safe/43393252#43393252
        self.model._make_predict_function()

    def _set_gen(self, mode):
        gen = self.data_store.get("generator", f"general.{mode}")
        setattr(self, f"{mode}_set", Distributor(gen, self.model, self.batch_size))

    def set_generators(self):
        for mode in ["train", "val", "test"]:
            self._set_gen(mode)

    def train(self):
        logging.info(f"Train with {len(self.train_set)} mini batches.")
        history = self.model.fit_generator(generator=self.train_set.distribute_on_batches(),
                                           steps_per_epoch=len(self.train_set),
                                           epochs=self.epochs,
                                           verbose=2,
                                           validation_data=self.val_set.distribute_on_batches(),
                                           validation_steps=len(self.val_set),
                                           callbacks=[self.checkpoint, self.lr_sc])
        self.save_callbacks(history)

    def save_callbacks(self, history):
        path = self.data_store.get("experiment_path", "general")
        with open(os.path.join(path, "history.json"), "w") as f:
            json.dump(history.history, f)
        with open(os.path.join(path, "history_lr.json"), "w") as f:
            json.dump(self.lr_sc.lr, f)



